/* This file is in the public domain */
#include <iostream>
#include <fstream>
#include <cctype>
#include <string>
#include <vector>

#include <opencl/rsa.h>
#include <opencl/dsa.h>
#include <opencl/rw.h>
#include <opencl/nr.h>
#include <opencl/dh.h>

#include <opencl/look_pk.h>

#include <opencl/randpool.h>
#include <opencl/rng.h>
using namespace OpenCL;

SecureVector<byte> decode_hex(const std::string&);
BigInt to_bigint(const std::string&);

#define DEBUG 0

class Fixed_Output_RNG : public RandomNumberGenerator
   {
   public:
      byte random()
         {
         if(position < output.size())
            return output[position++];
         else
            return 42;
         }
      void clear() throw() {}
      std::string name() const { return "Fixed Output RNG"; }
      void add_randomness(const byte[], u32bit) throw() {}
      Fixed_Output_RNG(const SecureVector<byte>& x)
         {
         output = x;
         position = 0;
         }
   private:
      SecureVector<byte> output;
      u32bit position;
   };

std::vector<std::string> parse(const std::string&);
void strip(std::string&);

u32bit validate_dsa_sig(const std::string&, const std::vector<std::string>&);
u32bit validate_rsa_enc(const std::string&, const std::vector<std::string>&);
u32bit validate_rsa_sig(const std::string&, const std::vector<std::string>&);
u32bit validate_rw_sig(const std::string&, const std::vector<std::string>&);
u32bit validate_nr_sig(const std::string&, const std::vector<std::string>&);
u32bit validate_dh(const std::string&, const std::vector<std::string>&);

u32bit do_pk_validation_tests(const std::string& filename)
   {
   std::ifstream test_data(filename.c_str());

   if(!test_data)
       {
       std::cout << "Couldn't open test file " << filename << std::endl;
       std::exit(1);
       }

   u32bit errors = 0, alg_count = 0;
   std::string algorithm;

   while(!test_data.eof())
      {
      if(test_data.bad() || test_data.fail())
         {
         std::cout << "File I/O error." << std::endl;
         std::exit(1);
         }

      std::string line;
      std::getline(test_data, line);

      strip(line);
      if(line.size() == 0) continue;

      // Do line continuation
      while(line[line.size()-1] == '\\' && !test_data.eof())
         {
         line.replace(line.size()-1, 1, "");
         std::string nextline;
         std::getline(test_data, nextline);
         strip(nextline);
         if(nextline.size() == 0) continue;
         line += nextline;
         }

      if(line[0] == '[' && line[line.size() - 1] == ']')
         {
         algorithm = line.substr(1, line.size() - 2);
#if DEBUG
         std::cout << "Testing " << algorithm << std::endl;
#endif
         alg_count = 0;
         continue;
         }

      std::vector<std::string> substr = parse(line);

#if DEBUG
      std::cout << "Testing: " << algorithm << std::endl;
#endif

      u32bit new_errors = 0;

      if(algorithm.find("DSA_") != std::string::npos)
         new_errors = validate_dsa_sig(algorithm, substr);
      if(algorithm.find("NR_SHA1") != std::string::npos)
         new_errors = validate_nr_sig(algorithm, substr);
      if(algorithm.find("RSAES_") != std::string::npos)
         new_errors = validate_rsa_enc(algorithm, substr);
      if(algorithm.find("RSASSA_") != std::string::npos)
         new_errors = validate_rsa_sig(algorithm, substr);
      if(algorithm.find("RWSSA_") != std::string::npos)
         new_errors = validate_rw_sig(algorithm, substr);
      if(algorithm.find("DHKAS_") != std::string::npos)
         new_errors = validate_dh(algorithm, substr);

      alg_count++;
      errors += new_errors;

      if(new_errors)
         std::cout << "ERROR: \"" << algorithm << "\" failed test #"
                   << std::dec << alg_count << std::endl;
      }

   // reset the global rng, because the validate_xxx will have replaced it
   set_global_rng(new Randpool);

   return errors;
   }

void dump_data(const SecureVector<byte>& out,
               const SecureVector<byte>& expected)
   {
   // For whatever reason, I cannot figure out how to do this with cout...
   std::printf("Got: ");
   for(u32bit j = 0; j != out.size(); j++)
      std::printf("%02X", out[j]);

   std::printf("\nExp: ");
   for(u32bit j = 0; j != expected.size(); j++)
      std::printf("%02X", expected[j]);
   std::printf("\n");
   std::fflush(stdout);
   }

void validate_encryption(PK_Encryptor* e, PK_Decryptor* d,
                         const std::string& algo, const std::string& input,
                         const std::string& random, const std::string& exp,
                         bool& failure)
   {
   SecureVector<byte> message = decode_hex(input);
   set_global_rng(new Fixed_Output_RNG(decode_hex(random)));

   SecureVector<byte> expected = decode_hex(exp);

   SecureVector<byte> out = e->encrypt(message);
   if(out != expected)
      {
      std::cout << "FAILED (encrypt): " << algo << std::endl;
      dump_data(out, expected);
      failure = true;
      }

   SecureVector<byte> decrypted = d->decrypt(out);
   if(decrypted != message)
      {
      std::cout << "FAILED (decrypt): " << algo << std::endl;
      failure = true;
      }

   delete e;
   delete d;
   }

void validate_signature(PK_Verifier* v, PK_Signer* s, const std::string& algo,
                        const std::string& input, const std::string& random,
                        const std::string& exp, bool& failure)
   {
   SecureVector<byte> message = decode_hex(input);
   set_global_rng(new Fixed_Output_RNG(decode_hex(random)));

   SecureVector<byte> expected = decode_hex(exp);

   SecureVector<byte> sig = s->sign_message(message, message.size());

   if(sig != expected)
      {
      std::cout << "FAILED (sign): " << algo << std::endl;
      dump_data(sig, expected);
      failure = true;
      }

   if(!v->verify_message(message, message.size(), sig, sig.size()))
      {
      std::cout << "FAILED (verify): " << algo << std::endl;
      failure = true;
      }

   /* This isn't a very thorough testing method, but it will hopefully
      catch any really horrible errors */
   sig[sig.size()-5]++;
   if(v->verify_message(message, message.size(), sig, sig.size()))
      {
      std::cout << "FAILED (accepted bad sig): " << algo << std::endl;
      failure = true;
      }

   delete v;
   delete s;
   }

void validate_kas(PK_Key_Agreement_Scheme* kas, const std::string& algo,
                  const std::string& pubkey, const std::string& output,
                  u32bit keylen, bool& failure)
   {
   SecureVector<byte> input = decode_hex(pubkey);
   SecureVector<byte> expected = decode_hex(output);

   SecureVector<byte> got = kas->derive_key(input, keylen).copy();

   if(got != expected)
      {
      std::cout << "FAILED (kas): " << algo << std::endl;
      dump_data(got, expected);
      failure = true;
      }

   delete kas;
   }

u32bit validate_rsa_enc(const std::string& algo,
                        const std::vector<std::string>& str)
   {
   RSA_PrivateKey key(to_bigint(str[1]), to_bigint(str[2]),
                      to_bigint(str[0]));

   PK_Encryptor* e = 0;
   PK_Decryptor* d = 0;
   if(algo == "RSAES_EME1_SHA1")
      {
      e = get_pk_encryptor(key, "EME1(SHA-1)");
      d = get_pk_decryptor(key, "EME1(SHA-1)");
      }
   else if(algo == "RSAES_PKCS1v15")
      {
      e = get_pk_encryptor(key, "PKCS1v15");
      d = get_pk_decryptor(key, "PKCS1v15");
      }
   else if(algo == "RSAES_RAW")
      {
      e = get_pk_encryptor(key, "Raw");
      d = get_pk_decryptor(key, "Raw");
      }
   else
      {
      std::cout << "WARNING: pk.cpp doesn't know about " << algo << std::endl;
      return 0;
      }

   bool failure = false;
   validate_encryption(e, d, algo, str[3], str[4], str[5], failure);
   return (failure ? 1 : 0);
   }

u32bit validate_rsa_sig(const std::string& algo,
                        const std::vector<std::string>& str)
   {
   if(str.size() < 6)
      throw Exception("Invalid input from pk_valid.dat");

   RSA_PrivateKey key(to_bigint(str[1]), to_bigint(str[2]),
                      to_bigint(str[0]));

   PK_Verifier* v = 0;
   PK_Signer* s = 0;
   if(algo == "RSASSA_EMSA2_SHA1")
      {
      v = get_pk_verifier(key, "EMSA2(SHA-1)");
      s = get_pk_signer(key, "EMSA2(SHA-1)");
      }
   else if(algo == "RSASSA_EMSA3_SHA1")
      {
      v = get_pk_verifier(key, "EMSA3(SHA-1)");
      s = get_pk_signer(key, "EMSA3(SHA-1)");
      }
   else if(algo == "RSASSA_EMSA3_MD5")
      {
      v = get_pk_verifier(key, "EMSA3(MD5)");
      s = get_pk_signer(key, "EMSA3(MD5)");
      }
   else if(algo == "RSASSA_RAW_SHA1")
      {
      v = get_pk_verifier(key, "EMSA1(SHA-1)");
      s = get_pk_signer(key, "EMSA1(SHA-1)");
      }
   else
      {
      std::cout << "WARNING: pk.cpp doesn't know about " << algo
                << std::endl;
      return 0;
      }

   bool failure = false;
   validate_signature(v, s, algo, str[3], str[4], str[5], failure);
   return (failure ? 1 : 0);
   }

u32bit validate_rw_sig(const std::string& algo,
                       const std::vector<std::string>& str)
   {
   if(str.size() < 6)
      throw Exception("Invalid input from pk_valid.dat");

   RW_PrivateKey key(to_bigint(str[1]), to_bigint(str[2]), to_bigint(str[0]));

   PK_Verifier* v = 0;
   PK_Signer* s = 0;
   if(algo == "RWSSA_EMSA2_SHA1")
      {
      v = get_pk_verifier(key, "EMSA2(SHA-1)");
      s = get_pk_signer(key, "EMSA2(SHA-1)");
      }
   else
      {
      std::cout << "WARNING: pk.cpp doesn't know about " << algo << std::endl;
      return 0;
      }

   bool failure = false;
   validate_signature(v, s, algo, str[3], str[4], str[5], failure);
   return (failure ? 1 : 0);
   }

u32bit validate_dsa_sig(const std::string& algo,
                        const std::vector<std::string>& str)
   {
   if(str.size() != 8)
      throw Exception("Invalid input from pk_valid.dat");

   DL_Group domain(to_bigint(str[0]), to_bigint(str[1]),
                               to_bigint(str[2]));
   DSA_PrivateKey key(domain, to_bigint(str[4]), to_bigint(str[3]));

   PK_Verifier* v = 0;
   PK_Signer* s = 0;
   if(algo == "DSA_SHA1")
      {
      v = get_pk_verifier(key, "EMSA1(SHA-1)");
      s = get_pk_signer(key, "EMSA1(SHA-1)");
      }
   else if(algo == "DSA_Tiger")
      {
      v = get_pk_verifier(key, "EMSA1(Tiger)");
      s = get_pk_signer(key, "EMSA1(Tiger)");
      }
   else
      {
      std::cout << "WARNING: pk.cpp doesn't know about " << algo << std::endl;
      return 0;
      }

   bool failure = false;
   validate_signature(v, s, algo, str[5], str[6], str[7], failure);
   return (failure ? 1 : 0);
   }

u32bit validate_nr_sig(const std::string& algo,
                       const std::vector<std::string>& str)
   {
   if(str.size() != 8)
      throw Exception("Invalid input from pk_valid.dat");

   DL_Group domain(to_bigint(str[0]), to_bigint(str[1]),
                               to_bigint(str[2]));

   NR_PrivateKey key(domain, to_bigint(str[4]),
                     to_bigint(str[3]));

   PK_Verifier* v = get_pk_verifier(key, "EMSA1(SHA-1)");
   PK_Signer* s = get_pk_signer(key, "EMSA1(SHA-1)");

   bool failure = false;
   validate_signature(v, s, algo, str[5], str[6], str[7], failure);
   return (failure ? 1 : 0);
   }

u32bit validate_dh(const std::string& algo,
                   const std::vector<std::string>& str)
   {
   if(str.size() != 5 && str.size() != 6)
      throw Exception("Invalid input from pk_valid.dat");

   DL_Group domain(to_bigint(str[0]), to_bigint(str[1]));

   DH_PrivateKey key(domain, to_bigint(str[2]));

   u32bit keylen = 0;
   if(str.size() == 6)
      keylen = std::atoi(str[5].c_str());

   PK_Key_Agreement_Scheme* kas;
   if(algo == "DHKAS_RAW")
      kas = get_pk_kas(key, "Raw");
   else if(algo == "DHKAS_KDF2_SHA1")
      kas = get_pk_kas(key, "KDF2(SHA-1)");
   else
      {
      std::cout << "WARNING: pk.cpp doesn't know about " << algo << std::endl;
      return 0;
      }

   bool failure = false;
   validate_kas(kas, algo, str[3], str[4], keylen, failure);
   return (failure ? 1 : 0);
   }
