/*
Grab an RSA public key from the file given as an argument, grab a message
from another file, and encrypt the message.

Algorithms used:
  RSA with EME1(SHA-1) padding to encrypt the master key
  CAST-128 in CBC mode with PKCS#7 padding to encrypt the message.
  HMAC with SHA-1 is used to authenticate the message

The keys used are derived from the master key (the thing that's encrypted with
RSA) using MD5. The master key is prefixed with some value P, where P is
"CAST", "IV" or "MAC", in order to parameterize the output for each key. The
keys for CAST-128 and HMAC are used as-is, while the output used for the IV is
truncated (the first half is used).

Note that the key format is just a pair of lines, first one is the exponent,
second one is the modulus. No provisions for anything complicated... that's why
this is only an example program.

Written by Jack Lloyd (lloyd@acm.jhu.edu), June 3, 2002

This file is in the public domain
*/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <string>

#include <opencl/pubkey.h>
#include <opencl/ieee1363.h>
#include <opencl/rsa.h>

#include <opencl/cast128.h>
#include <opencl/hmac.h>
#include <opencl/sha1.h>
#include <opencl/md5.h>

#include <opencl/pipe.h>
#include <opencl/filters.h>
#include <opencl/cbc.h>

#include <opencl/base64.h>
using namespace OpenCL_types;

std::string b64_encode(const byte[], u32bit);
OpenCL::SymmetricKey derive_key(const std::string&,
                                const OpenCL::SymmetricKey&, u32bit);

void write_bigint_to_pipe(OpenCL::Pipe&, const OpenCL::BigInt&);

int main(int argc, char* argv[])
   {
   try {

   if(argc != 3)
      {
      std::cout << "Usage: rsa_enc keyfile messagefile" << std::endl;
      return 1;
      }

   std::ifstream keyfile(argv[1]);
   if(!keyfile)
      {
      std::cout << "Couldn't read the key file." << std::endl;
      return 1;
      }

   std::ifstream message(argv[2]);
   if(!message)
      {
      std::cout << "Couldn't read the message file." << std::endl;
      return 1;
      }

   std::string output_name(argv[2]);
   output_name += ".enc";
   std::ofstream ciphertext(output_name.c_str());
   if(!ciphertext)
      {
      std::cout << "Couldn't write the ciphertext to " << output_name
           << std::endl;
      return 1;
      }

   OpenCL::LibraryInitializer init;

   std::string e, n;
   getline(keyfile, e);
   getline(keyfile, n);

   OpenCL::RSA_PublicKey key(n, e);
   key.check_params();

   OpenCL::PK_Encryptor encryptor(key, new OpenCL::EME1("SHA-1"));

   OpenCL::SymmetricKey masterkey(16);
   if(masterkey.length() > encryptor.maximum_input_size())
      {
      std::cout << "Sorry, this RSA key is too small to encrypt a session key "
         "that is " << masterkey.length() << " bytes long\n";
      return 1;
      }

   OpenCL::SymmetricKey cast_key = derive_key("CAST", masterkey, 16);
   OpenCL::SymmetricKey mac_key  = derive_key("MAC",  masterkey, 16);
   OpenCL::SymmetricKey iv       = derive_key("IV",   masterkey, 8);

   OpenCL::SecureVector<byte> encrypted_key =
      encryptor.encrypt(masterkey.value());

   ciphertext << b64_encode(encrypted_key, encrypted_key.size()) << std::endl;
   
   OpenCL::Pipe pipe(new OpenCL::Fork(
                new OpenCL::Chain(
                   new OpenCL::CBC_Encryption(new OpenCL::CAST_128,
                                              new OpenCL::PKCS7_Padding,
                                              cast_key, iv),
                   new OpenCL::Base64_Encoder(true)
                   ),
                new OpenCL::Chain(
                   new OpenCL::MAC_Filter(new OpenCL::HMAC("SHA-1"),
                                          mac_key, 12),
                   new OpenCL::Base64_Encoder
                   )
                )
      );

   pipe.start_msg();
   message >> pipe;
   pipe.end_msg();

   /* Write the MAC as the second line. That way we can pull it off right from
      the start, and feed the rest of the file right into a pipe on the 
      decrypting end.
   */
   ciphertext << pipe.read_all_as_string(1) << std::endl;
   ciphertext << pipe.read_all_as_string(0);

   }

   catch(OpenCL::Exception& e)
      {
      std::cout << "Exception caught: " << e.what() << std::endl;
      }
   catch(std::exception& e)
      {
      std::cout << "Standard library exception caught: " << e.what()
                << std::endl;
      }
   catch(...)
      {
      std::cout << "Unknown exception caught." << std::endl;
      }

   return 0;
   }

/*
  Maybe this should be in the library directly?

  Hrm... it's really not that general. But at the same time general enough
  that I do use it fairly often.
*/
std::string b64_encode(const byte in[], u32bit length)
   {
   OpenCL::Pipe pipe(new OpenCL::Base64_Encoder);
   pipe.start_msg();
   pipe.write(in, length);
   pipe.end_msg();

   return pipe.read_all_as_string();
   }

OpenCL::SymmetricKey derive_key(const std::string& param,
                        const OpenCL::SymmetricKey& masterkey,
                        u32bit outputlength)
   {
   if(outputlength > 16)
      throw OpenCL::Exception("Can't derive a key longer than 16 bytes "
                              "with MD5");

   OpenCL::MD5 md5;
   md5.update(param);
   md5.update(masterkey, masterkey.length());
   OpenCL::SecureVector<byte> hash = md5.final();

   return OpenCL::SymmetricKey(hash, outputlength);
   }

