/*
Generate an RSA key of a specified bitlength, and put it into a file. The
public key is written in plaintext as a exponent/modulus pair, the private key
is encrypted using CAST-128 in CBC mode, and MACed with HMAC-SHA1-96. For the
private key, we just store the modulus and the pair of primes, since everything
else is easily derivable from them.

The keys/ivs are derived using OpenPGP's S2K algorithm, with an 8 byte salt
(which is written as the first 8 bytes of the private key file), and an
iteration count of 8192. The passphrase is prefixed with a string to make sure
that each one is different, and hopefully not derivable from one another.

The CAST key uses a prefix of "CAST"
The HMAC key uses a prefix of "MAC"
The IV uses a prefix of "IV"

Written by Jack Lloyd (lloyd@acm.jhu.edu), June 2-3, 2002

This file is in the public domain
*/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <string>

#include <opencl/rsa.h>
#include <opencl/cast128.h>
#include <opencl/hmac.h>
#include <opencl/sha1.h>
#include <opencl/pgp_s2k.h>

#include <opencl/pipe.h>
#include <opencl/filters.h>
#include <opencl/cbc.h>

#include <opencl/base64.h>
using namespace OpenCL_types;

std::string b64_encode(const byte[], u32bit);
void write_bigint_to_pipe(OpenCL::Pipe&, const OpenCL::BigInt&);

int main(int argc, char* argv[])
   {
   try {

   if(argc != 3)
      {
      std::cout << "Usage: rsa_kgen bitsize passphrase" << std::endl;
      return 1;
      }

   u32bit bits = std::atoi(argv[1]);
   if(bits < 512 || bits > 4096)
      {
      std::cout << "Invalid argument for bitsize" << std::endl;
      return 1;
      }

   std::string passphrase(argv[2]);

   std::ofstream priv("rsakey.priv");
   std::ofstream pub("rsakey.pub");
   if(!priv || !pub)
      {
      std::cout << "Couldn't write output files" << std::endl;
      return 1;
      }

   OpenCL::LibraryInitializer init;
   OpenCL::RSA_PrivateKey key(bits);

   pub << key.get_e() << "\n" << key.get_n() << std::endl;

   OpenCL::OpenPGP_S2K s2k("SHA-1", 8192, 8);

   OpenCL::BlockCipher_Key cast_key = s2k.derive_key("CAST" + passphrase, 16);
   OpenCL::BlockCipherModeIV iv = s2k.derive_key("IV" + passphrase, 8);
   OpenCL::MAC_Key mac_key = s2k.derive_key("MAC" + passphrase, 16);

   OpenCL::SecureVector<byte> salt = s2k.current_salt();
   priv << b64_encode(salt, salt.size()) << std::endl;
   
   OpenCL::Pipe pipe(new OpenCL::Fork(
                new OpenCL::Chain(
                   new OpenCL::CBC_Encryption(new OpenCL::CAST_128,
                                              new OpenCL::PKCS7_Padding,
                                              cast_key, iv),
                   new OpenCL::Base64_Encoder(true)
                   ),
                new OpenCL::Chain(
                   new OpenCL::MAC_Filter(new OpenCL::HMAC("SHA-1"),
                                          mac_key, 12),
                   new OpenCL::Base64_Encoder
                   )
                )
      );

   pipe.start_msg();
   write_bigint_to_pipe(pipe, key.get_e());
   write_bigint_to_pipe(pipe, key.get_p());
   write_bigint_to_pipe(pipe, key.get_q());
   pipe.end_msg();

   priv << pipe.read_all_as_string(1) << std::endl;
   priv << pipe.read_all_as_string(0);

   }

   catch(OpenCL::Exception& e)
      {
      std::cout << "Exception caught: " << e.what() << std::endl;
      }
   catch(std::exception& e)
      {
      std::cout << "Standard library exception caught: " << e.what()
                << std::endl;
      }
   catch(...)
      {
      std::cout << "Unknown exception caught." << std::endl;
      }

   return 0;
   }

std::string b64_encode(const byte in[], u32bit length)
   {
   OpenCL::Pipe pipe(new OpenCL::Base64_Encoder);
   pipe.start_msg();
   pipe.write(in, length);
   pipe.end_msg();

   return pipe.read_all_as_string();
   }

/* Write the length as a 4 byte big endian value, followed up by the
   value itself
*/
void write_bigint_to_pipe(OpenCL::Pipe& pipe, const OpenCL::BigInt& num)
   {
   OpenCL::SecureVector<byte> bits = num.binary_encode();

   for(u32bit j = 0; j != 4; j++)
      pipe.write(OpenCL::get_byte(j, bits.size()));
   pipe.write(bits, bits.size());
   }
