/*************************************************
* Initalization Function Source File             *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/init.h>
#include <opencl/allocate.h>
#include <opencl/rng.h>
#include <opencl/mutex.h>
#include <opencl/timers.h>
#include <opencl/randpool.h>
#include <opencl/x917_rng.h>
#include <opencl/look_add.h>
#include <opencl/es_file.h>
#include <map>

#if defined(OPENCL_EXT_MUTEX_PTHREAD)
  #include <opencl/mux_pthr.h>
#endif

#if defined(OPENCL_EXT_ENTROPY_SRC_UNIX)
  #include <opencl/es_unix.h>
#endif

#if defined(OPENCL_EXT_ENTROPY_SRC_FTW)
  #include <opencl/es_ftw.h>
#endif

#if defined(OPENCL_EXT_ENTROPY_SRC_EGD)
  #include <opencl/es_egd.h>
#endif

#if defined(OPENCL_EXT_ALLOC_MMAP)
  #include <opencl/mmap_mem.h>
#endif

#if defined(OPENCL_EXT_ALLOC_MLOCK)
  #include <opencl/lock_mem.h>
#endif

#if defined(OPENCL_EXT_TIMER_HARDWARE)
  #include <opencl/tm_hard.h>
#elif defined(OPENCL_EXT_TIMER_POSIX)
  #include <opencl/tm_posix.h>
#elif defined(OPENCL_EXT_TIMER_UNIX)
  #include <opencl/tm_unix.h>
#endif

namespace OpenCL {

namespace {

/*************************************************
* Parse the option string                        *
*************************************************/
std::map<std::string, std::string> parse_args(const std::string& flags)
   {
   std::string opt_string = flags + " ";
   std::map<std::string, std::string> options;

   std::string::size_type start = 0, end = opt_string.find(" ");
   while(end != std::string::npos)
      {
      std::string option = opt_string.substr(start, end-start);
      start = end+1;
      end = opt_string.find(" ", start);

      if(option == "")
         continue;

      if(option.find("=") == std::string::npos)
         options[option] = "";
      else
         {
         std::string::size_type split = option.find("=");

         std::string opt_name = option.substr(0, split);
         std::string arg = option.substr(split + 1, option.length());

         if(opt_name == "")
            continue;

         options[opt_name] = arg;
         }
      }

   return options;
   }

/*************************************************
* Check if an option is set in the argument      *
*************************************************/
bool arg_set(const std::map<std::string, std::string>& args,
             const std::string& option)
   {
   if(args.find(option) != args.end())
      return true;
   return false;
   }

/*************************************************
* Register a mutex type, if possible             *
*************************************************/
bool set_mutex()
   {
#if defined(OPENCL_EXT_MUTEX_PTHREAD)
   set_mutex_type(new Pthread_Mutex);
   return true;
#else
   return false;
#endif
   }

/*************************************************
* Register a high resolution timer, if possible  *
*************************************************/
void set_timer()
   {
#if defined(OPENCL_EXT_TIMER_HARDWARE)
   set_timer_type(new Hardware_Timer);
#elif defined(OPENCL_EXT_TIMER_POSIX)
   set_timer_type(new POSIX_Timer);
#elif defined(OPENCL_EXT_TIMER_UNIX)
   set_timer_type(new Unix_Timer);
#endif
   }

/*************************************************
* Register a more secure allocator, if possible  *
*************************************************/
void set_safe_allocator()
   {
#if defined(OPENCL_EXT_ALLOC_MLOCK)
   add_allocator_type("mlock", new MemoryLocking_Allocator);
   set_default_allocator("mlock");
#endif

#if defined(OPENCL_EXT_ALLOC_MMAP)
   add_allocator_type("mmap", new MemoryMapping_Allocator);
   set_default_allocator("mmap");
#endif
   }

/*************************************************
* Add entropy to the global RNG                  *
*************************************************/
void seed_rng(bool several_sources)
   {
   File_EntropySource es_file;
   Global_RNG::add_entropy(es_file);

   if(!several_sources)
      return;

#if defined(OPENCL_EXT_ENTROPY_SRC_UNIX)
   Unix_EntropySource es_unix;
   Global_RNG::add_entropy(es_unix);
#endif

#if defined(OPENCL_EXT_ENTROPY_SRC_FTW)
   FTW_EntropySource es_ftw;
   Global_RNG::add_entropy(es_ftw);
#endif
   }

/*************************************************
* Add entropy from EGD to the global RNG         *
*************************************************/
void seed_rng_egd(const std::string egd_path)
   {
#if defined(OPENCL_EXT_ENTROPY_SRC_EGD)
   EGD_EntropySource es_egd(egd_path);
   Global_RNG::add_entropy(es_egd);
#endif
   }

}

/*************************************************
* Internal startup/shutdown functions            *
*************************************************/
extern void initalize_memory_subsystem();
extern void shutdown_memory_subsystem();
extern void create_global_bigints();
extern void delete_global_bigints();
extern void load_algorithms_list();

/*************************************************
* Libary Initalization                           *
*************************************************/
LibraryInitializer::LibraryInitializer(const std::string& arg_string)
   {
   std::map<std::string, std::string> args = parse_args(arg_string);

   if(arg_set(args, "thread_safe"))
      if(!set_mutex())
         throw Exception("LibraryInitializer: thread safety impossible");

   initalize_memory_subsystem();
   create_global_bigints();
   load_algorithms_list();

   if(arg_set(args, "secure_memory"))
      set_safe_allocator();

   if(!arg_set(args, "no_timers"))
      set_timer();

   if(arg_set(args, "fast_rng"))
      set_global_rng(new ANSI_X917_RNG);
   else
      set_global_rng(new Randpool);

   seed_rng(arg_set(args, "seed_rng"));

   if(arg_set(args, "egd_path"))
      seed_rng_egd(args["egd_path"]);
   }

/*************************************************
* Libary Deinitialization                        *
*************************************************/
LibraryInitializer::~LibraryInitializer()
   {
   set_global_rng(0);
   destroy_lookup_tables();
   delete_global_bigints();
   set_timer_type(0);
   set_mutex_type(0);
   shutdown_memory_subsystem();
   }

}
