/*************************************************
* Algorithm Lookup Table Source File             *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/lookup.h>
#include <opencl/look_add.h>
#include <map>

namespace OpenCL {

namespace {

std::map<std::string, BlockCipher*> bc_map;
std::map<std::string, StreamCipher*> sc_map;
std::map<std::string, HashFunction*> hf_map;
std::map<std::string, MessageAuthenticationCode*> mac_map;
std::map<std::string, BlockCipherModePaddingMethod*> bc_mode_pad_map;

std::map<std::string, std::string> alias_map;

}

/*************************************************
* Retrieve a block cipher                        *
*************************************************/
const BlockCipher* retrieve_block_cipher(const std::string& name)
   {
   std::map<std::string, BlockCipher*>::const_iterator algo;
   algo = bc_map.find(deref_alias(name));
   if(algo == bc_map.end())
      return 0;
   return algo->second;
   }

/*************************************************
* Retrieve a stream cipher                       *
*************************************************/
const StreamCipher* retrieve_stream_cipher(const std::string& name)
   {
   std::map<std::string, StreamCipher*>::const_iterator algo;
   algo = sc_map.find(deref_alias(name));
   if(algo == sc_map.end())
      return 0;
   return algo->second;
   }

/*************************************************
* Retrieve a hash function                       *
*************************************************/
const HashFunction* retrieve_hash(const std::string& name)
   {
   std::map<std::string, HashFunction*>::const_iterator algo;
   algo = hf_map.find(deref_alias(name));
   if(algo == hf_map.end())
      return 0;
   return algo->second;
   }

/*************************************************
* Retrieve a MAC                                 *
*************************************************/
const MessageAuthenticationCode* retrieve_mac(const std::string& name)
   {
   std::map<std::string, MessageAuthenticationCode*>::const_iterator algo;
   algo = mac_map.find(deref_alias(name));
   if(algo == mac_map.end())
      return 0;
   return algo->second;
   }

/*************************************************
* Retrieve a block cipher padding method         *
*************************************************/
const BlockCipherModePaddingMethod* retrieve_bc_pad(const std::string& name)
   {
   std::map<std::string, BlockCipherModePaddingMethod*>::const_iterator algo;
   algo = bc_mode_pad_map.find(deref_alias(name));
   if(algo == bc_mode_pad_map.end())
      return 0;
   return algo->second;
   }

/*************************************************
* Add a block cipher to the lookup table         *
*************************************************/
void add_algorithm(BlockCipher* algo)
   {
   if(bc_map.find(algo->name()) != bc_map.end())
      delete bc_map[algo->name()];
   bc_map[algo->name()] = algo;
   }

/*************************************************
* Add a stream cipher to the lookup table        *
*************************************************/
void add_algorithm(StreamCipher* algo)
   {
   if(sc_map.find(algo->name()) != sc_map.end())
      delete sc_map[algo->name()];
   sc_map[algo->name()] = algo;
   }

/*************************************************
* Add a hash function to the lookup table        *
*************************************************/
void add_algorithm(HashFunction* algo)
   {
   if(hf_map.find(algo->name()) != hf_map.end())
      delete hf_map[algo->name()];
   hf_map[algo->name()] = algo;
   }

/*************************************************
* Add a MAC to the lookup table                  *
*************************************************/
void add_algorithm(MessageAuthenticationCode* algo)
   {
   if(mac_map.find(algo->name()) != mac_map.end())
      delete mac_map[algo->name()];
   mac_map[algo->name()] = algo;
   }

/*************************************************
* Add a padding method to the lookup table       *
*************************************************/
void add_algorithm(BlockCipherModePaddingMethod* algo)
   {
   if(bc_mode_pad_map.find(algo->name()) != bc_mode_pad_map.end())
      delete bc_mode_pad_map[algo->name()];
   bc_mode_pad_map[algo->name()] = algo;
   }

/*************************************************
* Add an alias for an algorithm                  *
*************************************************/
void add_alias(const std::string& alias, const std::string& official_name)
   {
   if(alias_map.find(alias) != alias_map.end())
      throw Invalid_Argument("add_alias: The alias " + alias +
                             " already exists");
   alias_map[alias] = official_name;
   }

/*************************************************
* Dereference an alias                           *
*************************************************/
std::string deref_alias(const std::string& name)
   {
   std::map<std::string, std::string>::const_iterator realname;
   realname = alias_map.find(name);
   if(realname == alias_map.end())
      return name;
   return deref_alias(realname->second);
   }

/*************************************************
* Destroy the lookup tables                      *
*************************************************/
void destroy_lookup_tables()
   {
   std::map<std::string, BlockCipher*>::iterator bc_iter;
   for(bc_iter = bc_map.begin(); bc_iter != bc_map.end(); bc_iter++)
      delete bc_iter->second;

   std::map<std::string, StreamCipher*>::iterator sc_iter;
   for(sc_iter = sc_map.begin(); sc_iter != sc_map.end(); sc_iter++)
      delete sc_iter->second;

   std::map<std::string, HashFunction*>::iterator hf_iter;
   for(hf_iter = hf_map.begin(); hf_iter != hf_map.end(); hf_iter++)
      delete hf_iter->second;

   std::map<std::string, MessageAuthenticationCode*>::iterator mac_iter;
   for(mac_iter = mac_map.begin(); mac_iter != mac_map.end(); mac_iter++)
      delete mac_iter->second;

   bc_map.clear();
   sc_map.clear();
   hf_map.clear();
   mac_map.clear();
   }

}
